from typing import Any, Callable, Dict, Tuple

import jax.numpy as jnp
from jax import jacrev, jit, lax, ops, partial, random, vmap, tree_map
from jax.scipy.stats import norm

from .gp import GParameters, predict, train


Array = Any


def jacobian(f: Callable) -> Callable:
    return jit(jacrev(f))


def expected_improvement(
    x_pred: Array,
    params: GParameters,
    x: Array,
    y: Array,
    xi: float = 0.01,
) -> Array:
    """
    Computes the expected improvement at points x_pred over a
    Gaussian process trained on x and y.

    Parameters:
    -----------
    x_pred: The points at which the improvement is computed.
    params: Trained hyperparameters of the GP.
    x: Sampled points.
    y: Target values of the sampled points.
    xi: Parameter to balance exploration-exploitation.

    Returns:
    --------
    ei: The expected improvement of x_pred on the trained GP.
    """
    y_max = y.max()
    mu, std = predict(params, x, y, xt=x_pred)
    improvement = mu.T - y_max - xi
    z = improvement / (std + 1e-3)
    ei = improvement * norm.cdf(z) + std * norm.pdf(z)
    return ei


def replace_nan_values(arr: Array) -> Array:
    """
    Replaces the NaN values (if any) in arr with 0.

    Parameters:
    -----------
    arr: The array where NaN is removed from.

    Returns:
    --------
    The array with all the NaN elements replaced with 0.
    """
    # todo(alonfnt): Find a more robust solution.
    return jnp.where(jnp.isnan(arr), 0, arr)


@jit
def suggest_next(
    key: Array,
    params: GParameters,
    x: Array,
    y: Array,
    bounds: Array,
    n_seed: int = 1000,
    lr: float = 0.1,
    n_epochs: int = 150,
) -> Tuple[Array, Array]:
    """
    Suggests the new point to sample by optimizing the acquisition function.

    Parameters:
    -----------
    key: The pseudo-random generator key used for jax random functions.
    params: Hyperparameters of the Gaussian Process Regressor.
    x: Sampled points.
    y: Sampled targets.
    bounds: Array of (2, dim) shape with the lower and upper bounds of the
            variables.y_max: The current maximum value of the target values Y.
    n_seed (optional): the number of points to probe and minimize until
            finding the one that maximizes the acquisition functions.
    lr (optional): The step size of the gradient descent.
    n_epochs (optional): The number of steps done on the descent to minimize
            the seeds.


    Returns:
    --------
    A tuple with the parameters that maximize the acquisition function and a
    jax PRGKey to be used in the next sampling.
    """

    key1, key2 = random.split(key, 2)
    dim = bounds.shape[0]

    domain = random.uniform(
        key1, shape=(n_seed, dim), minval=bounds[:, 0], maxval=bounds[:, 1]
    )

    acq = partial(expected_improvement, params=params, x=x, y=y)

    J = jacobian(lambda x: acq(x.reshape(-1, dim)).reshape())
    HS = vmap(lambda x: x + lr * J(x))

    domain = lax.fori_loop(0, n_epochs, lambda _, d: HS(d), domain)
    domain = jnp.clip(
        domain.reshape(-1, dim), a_min=bounds[:, 0], a_max=bounds[:, 1]
    )
    domain = replace_nan_values(domain)

    ys = acq(domain)
    next_X = domain[ys.argmax()]
    return next_X, key2


def optim(
    f: Callable,
    constrains: Dict,
    seed: int = 42,
    n_init: int = 5,
    n: int = 10,
    xi: float = 0.01,
):
    """
    Finds the inputs of 'f' that yield the maximum value between the given
    'constrains', after 'n_init' + 'n' iterations.

    Parameters:
    -----------
    f: Function to optimize.
    constrains: Dictionary with the domain of each input variable.
    seed: Pseudo-random number generator seed for reproducibility
    n_init: Number of initial evaluations before suggesting optimized samples.
    n: Number of sampling iterations.
    xi: Parameter to balance exploration-exploitation.

    Returns:
    --------
    The parameters that maximize the given 'f'.
    """

    assert n > 0, "Num of iterations n should be a positive integer"

    key = random.PRNGKey(seed)
    dim = len(constrains)
    bounds = jnp.asarray(list(constrains.values()))
    X = random.uniform(
        key,
        shape=(n_init, dim),
        minval=bounds[:, 0],
        maxval=bounds[:, 1],
    )
    Y = vmap(f)(*X.T)

    # Create an empty array with the same values to not perjudicate the gp
    X = jnp.pad(X, ((0, n), (0, 0)), mode="edge")
    Y = jnp.pad(Y, ((0, n)), mode="edge")

    # Initialize the GP parameters
    params = GParameters(
        noise=jnp.zeros((1, 1)) - 5.0,
        amplitude=jnp.zeros((1, 1)),
        lengthscale=jnp.zeros((1, 1)),
    )
    momentums = tree_map(lambda x: x * 0, params)
    scales = tree_map(lambda x: x * 0 + 1, params)

    # todo(alonfnt): make it JAX friendly
    for i in range(n):
        idx = n_init + i
        params, momentums, scales = train(
            x=X, y=Y, params=params, momentums=momentums, scales=scales
        )
        max_params, key = suggest_next(
            key,
            params,
            X,
            Y,
            bounds,
        )
        print(f"New max point: {max_params} and {f(*max_params)}")
        X = ops.index_update(X, ops.index[idx, ...], max_params)
        Y = ops.index_update(Y, ops.index[idx], f(*max_params))
    xmax = X[Y.argmax()]
    return xmax



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bayex']

package_data = \
{'': ['*']}

install_requires = \
['jax>=0.2.18,<0.3.0', 'jaxlib>=0.1.69,<0.2.0']

setup_kwargs = {
    'name': 'bayex',
    'version': '0.1.0a0',
    'description': 'Bayesian Optimization with Gaussian Processes powered by JAX',
    'long_description': "# BAYEX: Bayesian Optimization powered by JAX\nBayex is a Bayesian global optimization library using Gaussian processes.\nIn contrast to existing Bayesian optimization libraries, Bayex is designed for JAX.\n\nInstead of relaying on external libraries, Bayex only relies on JAX and its custom implementations, without requiring importing massive libraries such as `sklearn`.\n\n## What is Bayesian Optimization?\n\nBayesian Optimization (BO) methods are useful for optimizing functions that are expensive to evaluate, lack an analytical expression and whose evaluations can be contaminated by noise. These methods rely on a probabilistic model of the objective function, typically a Gaussian process (GP), upon which an acquisition function is built. The acquisition function guides the optimization process and measures the expected utility of performing an evaluation of the objective at a new point. \n\n## Why JAX?\nUsing JAX as a backend removes some of the limitations found on Python, as it gives us direct mapping to the XLA compiler.\n\nXLA compiles and runs the JAX code into several architectures such as CPU, GPU and TPU without hassle. But the device agnostic approach is not the reason to back XLA for future scientific programs. XLA provides with optimizations under the hood such as Just-In-Time compilation and automatic parallelization that make Python (with a NumPy-like approach) a suitable candidate on some High Performance Computing scenarios.\n\nAdditionally, JAX provides Python code with automatic differentiation, which helps identify the conditions that maximize the acquisition function.\n\n\n## Installation\nBayex can be installed using PyPI via `pip`:\n```\npip install bayex\n```\nor from GitHub directly\n```\npip install git+git://github.com/alonfnt/bayex.git\n```\n## Getting Started\n```python\nimport bayex\n\ndef f(x, y):\n    return -y ** 2 - (x - y) ** 2 + 3 * x / y - 2\n\nconstrains = {'x': (-10, 10), 'y': (0, 10)}\nx_max, y_max = bayex.optim(f, constrains=constrains, seed=42)\n```\n\n## Contributing\nEveryone can contribute to Bayex and we welcome pull requests as well as raised issues.\nIn order to contribute code, one must begin by forking the repository. This creates a copy of the repository on your account.\n\nBayex uses poetry as a packaging and dependency manager. Hence, once you have cloned your repo on your own machine, you can use\n```\npoetry install\n```\nto install on the dependencies needed.\nYou should start a new branch to write your changes on\n```\ngit checkout -b name-of-change\n``` \nor \n```\ngit branch name-of-change\ngit checkout name-of-change\n```\n\nIt is welcome if PR are composed of a single commit, to keep the feature <-> commit balance.\nPlease, when writing the commit message, follow the [conventional commit](https://www.conventionalcommits.org/en/v1.0.0/) specitifcation.\nOnce you have made your changes and created your commit. It is recommended to run the pre-commit checks.\n```\npre-commit run --all\n```\nas well as the tests to make sure everything works\n```\npytest -n auto tests/\n```\n\nRemember to amend your current commit with the fixes if any of the checks fails.\n\n## Planned Features\n- [ ] Optimization on continuos domains.\n- [ ] Integer parameters support.\n- [ ] Categorical Variables \n- [ ] Automatic Parallelization on XLA Devices.\n\n## Citation\nTo cite this repository\n```\n@software{\n  author = {Albert Alonso, Jacob Ungar Felding},\n  title = {{Bayex}: Bayesian Global Optimization with JAX tool for {P}ython+{N}um{P}y programs},\n  url = {http://github.com/alonfnt/bayex},\n  version = {0.0.0},\n  year = {2021},\n}\n```\n## References\n1. [A Tutorial on Bayesian Optimization](https://arxiv.org/abs/1807.02811)\n2. [BayesianOtpimization Library](https://github.com/fmfn/BayesianOptimization)\n3. [JAX: Autograd and XLA](https://github.com/google/jax)\n",
    'author': 'Albert Alonso',
    'author_email': 'alonfnt@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alonfnt/bayex',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
